/*
 * @(#)UnotOD.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

import java.util.*;

/**
 * UnotOD.java
 *
 * `OD' means `Occurrence Deliver'.
 *
 * Created: Sat Feb  5 03:04:00 2005
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $$
 * @see <a href="http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf">
 * Discovering Frequent Substructures in Large Unordered Trees (PDF)</a>
 */
public class UnotOD extends Unot
{

    /**
     * Creates a new <code>UnotOD</code> instance.
     *
     */
    protected UnotOD()
    {
        super();
    }


    /**
     * Creates a new <code>UnotOD</code> instance.
     *
     * @param docset a <code>DocSet</code> value
     */
    public UnotOD(DocSet docset)
    {
        super(docset);
    } // Unot constructor


    /**
     * Describe <code>expand</code> method here.
     *
     * @param pattern an <code>UnotPattern</code> value
     * @param occList an <code>OccList</code> value
     * @param minSup an <code>int</code> value
     */
    protected void expand(UnotPattern pattern, OccList occList, int minSup)
    {

        this.numberOfExpansion++;

        // 祳ɥۤƤкƵ
        if ((this.maxCodeSize != MAX_SIZE_NO_LIMIT)
            && (this.maxCodeSize < pattern.getCodeSize()))
            return;

        // ѽФǤСΥѥä
        if (this.docset.getOccSize(occList) < minSup) {
            return;
		}

		//ȤʤѽХѥ󤬸Ĥä
		maintainPatternInfo(pattern, occList);

		this.patset.add(pattern);

        int topDepth = pattern.getTopDepth();
        int copyDepth = pattern.getCopyDepth();
        List unotPatternList = new ArrayList();

        if (pattern.isBalanced(copyDepth)) { // if Lc == Rc
            for (int dep = 1; dep <= copyDepth+1; dep++) {
                int maxLabel = pattern.getNodeLabel(dep);

                OccDict bucket = new OccDict(); // OD
                bucket.occurrenceDeliver(pattern, occList, dep-1, maxLabel); // OD

                for (Iterator itr = bucket.keySet().iterator(); itr.hasNext(); ) {
                    Integer label = (Integer) itr.next();

                    UnotPattern newPattern = null;
                    if (label.intValue() != maxLabel) {
                        newPattern = pattern.rightmostExpand(pattern, new DLpair(dep, label.intValue()), dep);
                    } else if (dep <= topDepth) {
                        newPattern = pattern.rightmostExpand(pattern, new DLpair(dep, maxLabel), dep-1);
                    } else continue;

                    OccList newOccList = (OccList) bucket.get(label);
                    if (useBalanceCheck)
						newOccList = deleteNonCanonicalOcc(newPattern, newOccList, copyDepth, topDepth);
                    this.numberOfOccurrence += newOccList.size();
                    expand(newPattern, newOccList, minSup);
                }
            }

        } else { // if Lc != Rc
            /* `da node' means `disagreement node'. */
            int daNodeDepth = pattern.getDisagreeNodeDepth(copyDepth);
            int daNodeLabel = pattern.getDisagreeNodeLabel(copyDepth);

            // when ( 1 <= dep < daNodeDepth )
            for (int dep = 1; dep < daNodeDepth; dep++) {
                int maxLabel = pattern.getNodeLabel(dep);

                OccDict bucket = new OccDict(); // OD
                bucket.occurrenceDeliver(pattern, occList, dep-1, maxLabel); // OD

                for (Iterator itr = bucket.keySet().iterator(); itr.hasNext(); ) {
                    Integer label = (Integer) itr.next();

                    UnotPattern newPattern = null;
                    if (label.intValue() != maxLabel) {
                        newPattern = pattern.rightmostExpand(pattern, new DLpair(dep, label.intValue()), dep);
                    } else {
                        newPattern = pattern.rightmostExpand(pattern, new DLpair(dep, maxLabel), dep-1);
                    }

                    OccList newOccList = (OccList) bucket.get(label);
					if (useBalanceCheck)
						newOccList = deleteNonCanonicalOcc(newPattern, newOccList, copyDepth, topDepth);
                    this.numberOfOccurrence += newOccList.size();
                    expand(newPattern, newOccList, minSup);
                }

            }

            // when ( dep == daNodeDepth )
            OccDict bucket = new OccDict(); // OD
            bucket.occurrenceDeliver(pattern, occList, daNodeDepth-1, daNodeLabel); // OD

            for (Iterator itr = bucket.keySet().iterator(); itr.hasNext(); ) {
                Integer label = (Integer) itr.next();

                UnotPattern newPattern = null;
                if (label.intValue() != daNodeLabel) {
                    newPattern = pattern.rightmostExpand(pattern, new DLpair(daNodeDepth, label.intValue()), daNodeDepth);
                } else {
                    newPattern = pattern.rightmostExpand(pattern, new DLpair(daNodeDepth, daNodeLabel), copyDepth);
                }

                OccList newOccList = (OccList) bucket.get(label);
                if (useBalanceCheck)
					newOccList = deleteNonCanonicalOcc(newPattern, newOccList, copyDepth, topDepth);
                this.numberOfOccurrence += newOccList.size();
                expand(newPattern, newOccList, minSup);
            }

        }

    } // expand


    /**
     * Describe <code>deleteNonCanonicalOcc</code> method here.
     *
     * @param pattern an <code>UnotPattern</code> value
     * @param occList an <code>OccList</code> value
     * @param copyDepth an <code>int</code> value
     * @param topDepth an <code>int</code> value
     * @return an <code>OccList</code> value
     */
    protected OccList deleteNonCanonicalOcc(UnotPattern pattern, OccList occList,
                                             int copyDepth, int topDepth)
    {
        OccList newOccList = new OccList();

        for (int i = 0; i < occList.size(); i++) {
            Occurrence newOcc = (Occurrence) occList.get(i);
            if (isCanonicalOcc(pattern, newOcc, copyDepth, topDepth)) {
                newOccList.add(newOcc);
            }
        }

        return newOccList;
    }


    /**
     * Describe class <code>OccDict</code> here.
     *
     */
    private class OccDict extends HashMap
    {

        /**
         * Describe <code>get</code> method here.
         *
         * @param label an <code>int</code> value
         * @return an <code>OccList</code> value
         */
        public OccList get(int label)
        {
            return (OccList) this.get(new Integer(label));
        }


        /**
         * Describe <code>occurrenceDeliver</code> method here.
         *
         * @param pattern an <code>UnotPattern</code> value
         * @param occList an <code>OccList</code> value
         * @param depth an <code>int</code> value
         * @param maxLabel an <code>int</code> value
         */
        protected void occurrenceDeliver(UnotPattern pattern, OccList occList, int depth, int maxLabel)
        {

            int parentIndex = pattern.getNodeIndex(depth);

            for (int i = 0; i < occList.size(); i++) {

                Occurrence occ = (Occurrence) occList.get(i);
                Occurrence newOcc = null;

                UnotNode parent = (UnotNode) occ.get(parentIndex);
                NodeList children = parent.getChildren();

                next: for (int j = 0; j < children.size(); j++) {

                    UnotNode child = (UnotNode) children.get(j);
                    int topLabel = child.getLabelId();
                    if ((topLabel <= maxLabel) && ! occ.contains(child)){
                        newOcc = new Occurrence(occ, child);
                    } else continue;

                    Integer key = new Integer(topLabel);
                    if (this.containsKey(key)) {
                        OccList newOccList = (OccList) this.get(key);
                        newOccList.add(newOcc);
                    } else {
                        OccList newOccList = new OccList();
                        newOccList.add(newOcc);
                        this.put(key, newOccList);
                    }
                }
            }

        } // occurrenceDeliver

    } // OccDict


} // UnotOD
